package net.astah.stpa.stamp.notation.editor;

import static net.astah.dd.di.edit.provider.IItemGraphicalStyleProvider.Feature.FILL_COLOR;

import javax.swing.ImageIcon;

import com.google.common.collect.Iterables;

import net.astah.golf.draw.GIconButton;
import net.astah.golf.draw.GNode;
import net.astah.golf.draw.GRoundRect;
import net.astah.golf.geom2D.Pnt2d;
import net.astah.golf.util.ResourceBundleList;
import net.astah.golf.widget.CImageIcon;
import net.astah.golf.widget.swing.SwingIconRegistry;
import net.astah.golf.widget.swing.SwingImageIcon;
import net.astah.jomt.jsystem.PropertyManager;
import net.astah.jomt.jview.RectangleDrawSuggest;
import net.astah.notation.editor.GraphicalStyles;
import net.astah.notation.editor.IDiagramViewElement;
import net.astah.notation.editor.IMovableViewElement;
import net.astah.notation.editor.IResizableViewElement;
import net.astah.notation.editor.ITransformer;
import net.astah.notation.editor.IViewElement;
import net.astah.notation.editor.ViewElements;
import net.astah.notation.editor.adjuster.ResizingAdjuster;
import net.astah.stpa.stamp.notation.STAMPComponentShape;
import net.astah.stpa.stamp.provider.STAMPEditPlugin;

/**
 * コンポーネント要素の通常表記を表示するためのビューエレメントを実装するクラスです。
 */
public class STAMPComponentViewElement extends STAMPComponentBaseViewElement
        implements IMovableViewElement, IResizableViewElement {

    private RectangleDrawSuggest drawSuggest;
    private STAMPComponentHighlighter highlighter;
    private ResizingAdjuster.Factory resizingAdjusterFactory;

    private static final byte FRAME_COLOR = BLACK;
    private static final byte BODY_COLOR = -1;
    private static final byte KNOB_COLOR = BLACK;
    private static final byte SHADOW_COLOR = LIGHTGRAY;

    private static final double HIGHLIGHT_MARGIN = 5;

    public STAMPComponentViewElement(STAMPComponentShape model) {
        super(model);
    }

    @Override
    public <T> T getAdapter(Class<T> adapter) {
        if (adapter.isInstance(drawSuggest)) {
            return adapter.cast(drawSuggest);
        } else if (adapter.isInstance(resizingAdjusterFactory)) {
            return adapter.cast(resizingAdjusterFactory);
        } else if (adapter.isInstance(highlighter)) {
            return adapter.cast(highlighter);
        }
        return super.getAdapter(adapter);
    }

    @Override
    protected void init() {
        super.init();
        initHighlighter();
    }

    private void initHighlighter() {
        highlighter = new STAMPComponentHighlighter(this);
    }

    @Override
    protected void initDrawSuggest() {
        super.initDrawSuggest();

        drawSuggest = new STAMPComponentDrawSuggest(this);
        String key = "diagramview.button.drop_down_component";
        CImageIcon image = new SwingImageIcon((ImageIcon) SwingIconRegistry.INSTANCE
                .getIcon(STAMPEditPlugin.getPlugin().getImage("full/obj16/Link")));
        drawSuggest.addModeAction("net.astah.stpa.stamp.ui.control.mode.CreateControlLink", image,
                key);

        ResourceBundleList resource = PropertyManager.getJudeResourceBundle();
        if (resource != null
                && Boolean.valueOf(resource.getString("stpa.compartment_suggest"))) {
            CImageIcon compartmentImage = new SwingImageIcon((ImageIcon) SwingIconRegistry.INSTANCE
                    .getIcon(STAMPEditPlugin.getPlugin().getImage("full/suggest16/Compartment")));
            GIconButton insideDrawButton = new GIconButton(compartmentImage,
                    "net.astah.ui.control.CreateCompartment", 1);
            drawSuggest.addButton(insideDrawButton);
        }
    }

    @Override
    protected void initAdjusterFactory() {
        super.initAdjusterFactory();
        resizingAdjusterFactory = ViewElements.getAncestor(this, IDiagramViewElement.class)
                .getAdapter(ResizingAdjuster.Factory.class);
    }

    @Override
    protected void updateColor() {
        GraphicalStyles.applyFillColor(bodyRect, getModel(), BODY_COLOR);
        GraphicalStyles.applyStrokeColor(bodyRect, getModel(), FRAME_COLOR);

        bodyRect.setKnobColorNo(KNOB_COLOR);
        bodyRect.setShadowColorNo(SHADOW_COLOR);
    }

    @Override
    protected IViewElement getEditingTextViewElement() {
        return Iterables.getFirst(viewElements(STAMPLabelViewElement.class), null);
    }

    @Override
    public ITransformer startMove(Pnt2d startLocation) {
        return new STAMPComponentTransformer(this, startLocation);
    }

    @Override
    public ITransformer startResize(Pnt2d startLocation, int knob) {
        return new STAMPComponentTransformer(this, startLocation, knob);
    }

    @Override
    protected GNode createHighlightShape() {
        final GRoundRect highlightShape = createHighlightRect();
        highlightShape.setAlpha(0.5f);
        highlightShape.setFilled(true);
        highlightShape.setRimWidth(10);
        highlightShape
                .setForegroundColor(GraphicalStyles.getHighlightColor(getModel(), FILL_COLOR));
        highlight = highlightShape;
        return highlightShape;
    }

    private GRoundRect createHighlightRect() {
        double x = bodyRect.getX() - HIGHLIGHT_MARGIN;
        double y = bodyRect.getY() - HIGHLIGHT_MARGIN;
        double width = bodyRect.getWidth() + HIGHLIGHT_MARGIN * 2;
        double height = bodyRect.getHeight() + HIGHLIGHT_MARGIN * 2;
        return new GRoundRect(x, y, width, height);
    }

    @Override
    protected void layoutHighlightShape() {
        super.layoutHighlightShape();

        if (highlight != null) {
            GRoundRect rect = createHighlightRect();
            ((GRoundRect) highlight).setRoundRect(rect.getX(), rect.getY(), rect.getWidth(),
                    rect.getHeight());
        }
    }
}
