package net.astah.stpa.stamp.notation.util;

import static net.astah.stpa.stamp.notation.STAMPNotationPackage.Literals.STAMP_DIAGRAM_ELEMENT__STAMP_MODEL_ELEMENT;

import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EStructuralFeature.Setting;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.uml2.common.util.UML2Util;

import net.astah.dd.di.DiagramElement;

/**
 * STAMP 表記法モデルのユーティリティメソッドを提供します。
 */
public class STAMPDiagramElements {

    /**
     * 指定されたオブジェクトをモデル要素として参照する図要素モデルを検索して返します。
     * <p>
     * {@code ancestor} が未指定の場合は全ての図の図要素が対象になるので注意して下さい。
     * </p>
     * 
     * @param modelElement
     *            検索結果の図要素モデルがモデル要素として参照するオブジェクト
     * @param ancestor
     *            非 null 値の場合は、検索結果はこの図要素の配下に限定されます。
     *            典型的には、ある図の配下の図要素に限定するために図を指定します。
     * @param targetClass
     *            検索結果の図要素モデルを限定するためのクラス
     * @return 指定されたオブジェクトをモデル要素として参照する図要素モデルのリスト
     */
    public static <T extends DiagramElement> List<T> getDiagramElements(final EObject modelElement,
            final DiagramElement ancestor,
            final Class<T> targetClass) {
        if (modelElement == null) {
            return Collections.emptyList();
        }
        return UML2Util.getNonNavigableInverseReferences(modelElement).stream()
                .filter(setting -> isTargetSetting(setting, ancestor, targetClass))
                .map(setting -> targetClass.cast(setting.getEObject()))
                .collect(Collectors.toList());
    }

    /**
     * 指定されたオブジェクトをモデル要素として参照する図要素モデルを検索して返します。
     * <p>
     * {@code ancestor} が未指定の場合は全ての図の図要素が対象になるので注意して下さい。
     * </p>
     * 
     * @param modelElement
     *            検索結果の図要素モデルがモデル要素として参照するオブジェクト
     * @param ancestor
     *            非 null 値の場合は、検索結果はこの図要素の配下に限定されます。
     *            典型的には、ある図の配下の図要素に限定するために図を指定します。
     * @return 指定されたオブジェクトをモデル要素として参照する図要素モデルのリスト
     */
    public static List<DiagramElement> getDiagramElements(final EObject modelElement,
            final DiagramElement ancestor) {
        return getDiagramElements(modelElement, ancestor, DiagramElement.class);
    }

    private static <T> boolean isTargetSetting(final Setting setting,
            final DiagramElement ancestor,
            final Class<T> targetClass) {
        if (setting.getEStructuralFeature() != STAMP_DIAGRAM_ELEMENT__STAMP_MODEL_ELEMENT) {
            return false;
        }
        final EObject eObject = setting.getEObject();
        if (!targetClass.isInstance(eObject)) {
            return false;
        }
        if (ancestor != null && !EcoreUtil.isAncestor(ancestor, eObject)) {
            return false;
        }
        return true;
    }
}
