package net.astah.stpa.stamp.ui.control;

import java.util.Collections;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature.Setting;
import org.eclipse.emf.ecore.util.ECrossReferenceAdapter;
import org.eclipse.emf.edit.command.CommandParameter;
import org.eclipse.emf.edit.command.CreateChildCommand;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.emf.transaction.util.TransactionUtil;

import net.astah.dd.di.DiagramElement;
import net.astah.dd.di.util.DiagramElements;
import net.astah.emf.edit.job.CommandJob;
import net.astah.emf.edit.job.Jobs;
import net.astah.golf.widget.CCanvas;
import net.astah.jomt.jsystem.JomtGlobal;
import net.astah.jomt.jview.DiagramEditor;
import net.astah.jomt.jview.DiagramEditors;
import net.astah.notation.editor.ViewElements;
import net.astah.stpa.stamp.notation.STAMPNotationPackage;
import net.astah.stpa.stamp.ui.menu.handler.AddProcessValueHandler;
import net.astah.stpa.stamp.ui.menu.handler.AddProcessVariableHandler;

/**
 * プロセスモデル要素を作成するコマンドの基底クラスです。
 */
abstract class AddProcessModelElementCommand extends net.astah.golf.control.Command {
    private int index = CommandParameter.NO_INDEX;
    private DiagramElement ownerDiagramElement;
    private EObject createdProcessModelElement;

    @Override
    public void setArgumentString(final String arg) {
        index = Integer.valueOf(arg);
    }

    /**
     * 作成されるプロセスモデル要素を所有するモデルの図要素を設定します。
     * <p>
     * {@link AddProcessVariableHandler}や{@link AddProcessValueHandler}からの使用を想定しています。
     * 
     * @param ownerDiagramElement
     *            作成されるプロセスモデル要素を所有するモデルの図要素
     */
    public void setOwnerDiagramElement(final DiagramElement ownerDiagramElement) {
        this.ownerDiagramElement = ownerDiagramElement;
    }

    @Override
    public void execute() {
        if (ownerDiagramElement == null) {
            ownerDiagramElement = getOwnerDiagramElementBySelection();
        }
        final EObject ownerModel = DiagramElements.getModelElement(ownerDiagramElement);
        final TransactionalEditingDomain domain = TransactionUtil.getEditingDomain(ownerModel);
        final Command command = createCommand(domain, ownerModel, index);
        final CommandJob job = new CommandJob(domain, command);
        Jobs.toCompletableFuture(job, true).thenRun(this::postExecute);
        job.schedule();
    }

    private DiagramElement getOwnerDiagramElementBySelection() {
        final DiagramEditor editor = JomtGlobal.viewManager.getCurrentDiagramEditor();
        if (editor == null) {
            return null;
        }
        final DiagramElement[] selectedDiagramElements = editor.getSelectedPresentations();
        if (selectedDiagramElements == null) {
            return null;
        }
        for (final DiagramElement selectedDiagramElement : selectedDiagramElements) {
            final DiagramElement diagramElement = getOwnerDiagramElement(selectedDiagramElement);
            if (diagramElement != null) {
                return diagramElement;
            }
        }
        return null;
    }

    private Command createCommand(final TransactionalEditingDomain domain, final EObject ownerModel,
            int index) {
        createdProcessModelElement = createProcessModelElement();
        return CreateChildCommand.create(domain, ownerModel,
                new CommandParameter(ownerModel, getFeature(), createdProcessModelElement, index),
                Collections.singleton(ownerModel));
    }

    private void postExecute() {
        if (JomtGlobal.viewManager == null) {
            return;
        }
        final DiagramEditor currentDiagramEditor = JomtGlobal.viewManager.getCurrentDiagramEditor();
        if (currentDiagramEditor == null) {
            return;
        }
        final CCanvas canvas = currentDiagramEditor.getCanvas();
        if (canvas == null) {
            return;
        }
        final DiagramElement createdDiagramElement = getDiagramElementInTargetCompartment(
                createdProcessModelElement);
        if (createdDiagramElement == null) {
            return;
        }
        canvas.unselect();
        DiagramEditors.selectOn(currentDiagramEditor, createdDiagramElement);
        ViewElements.startEditing(canvas, createdDiagramElement);
    }

    private DiagramElement getDiagramElementInTargetCompartment(final EObject model) {
        final ECrossReferenceAdapter adapter = ECrossReferenceAdapter
                .getCrossReferenceAdapter(model);
        if (adapter == null) {
            return null;
        } else {
            final DiagramElement compartment = getCompartment(ownerDiagramElement);
            return adapter.getNonNavigableInverseReferences(model).stream()
                    .filter(s -> isDiagramElementReference(s))
                    .map(s -> (DiagramElement) s.getEObject())
                    .filter(d -> compartment.getOwnedElements().contains(d)).findFirst()
                    .orElse(null);
        }
    }

    private boolean isDiagramElementReference(final Setting setting) {
        return setting
                .getEStructuralFeature() == STAMPNotationPackage.Literals.STAMP_DIAGRAM_ELEMENT__STAMP_MODEL_ELEMENT;
    }

    /**
     * 作成されるプロセスモデル要素を所有するモデルの図要素を取得します。
     * 
     * @param selectedDiagramElement
     *            選択された図要素
     * @return 作成されるプロセスモデル要素を所有するモデルの図要素を取得します。
     */
    abstract DiagramElement getOwnerDiagramElement(final DiagramElement selectedDiagramElement);

    /**
     * プロセスモデル要素を作成します。
     * 
     * @return 作成されたプロセスモデル要素
     */
    abstract EObject createProcessModelElement();

    /**
     * 作成されたプロセスモデル要素が追加されるfeatureを取得します。
     * 
     * @return 作成されたプロセスモデル要素が追加されるfeature
     */
    abstract EReference getFeature();

    /**
     * 区画の図要素を取得します。
     * 
     * @param ownerDiagramElement
     *            作成されたプロセスモデル要素を所有するモデルの図要素
     * @return 区画の図要素
     */
    abstract DiagramElement getCompartment(final DiagramElement ownerDiagramElement);
}
