package net.astah.stpa.stamp.ui.control;

import static net.astah.emf.edit.command.DiagramCommands.createDiagramCommand;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import org.eclipse.emf.common.command.CompoundCommand;
import org.eclipse.emf.edit.command.CommandParameter;
import org.eclipse.emf.edit.domain.EditingDomain;

import com.google.common.collect.Collections2;

import net.astah.dd.dc.Bounds;
import net.astah.dd.dc.DCFactory;
import net.astah.dd.dc.Point;
import net.astah.dd.di.Diagram;
import net.astah.dd.di.DiagramElement;
import net.astah.dd.di.Edge;
import net.astah.dd.di.Shape;
import net.astah.dd.di.edit.command.DiagramContext;
import net.astah.dd.di.edit.command.SetEdgeEndCommand;
import net.astah.dd.di.edit.command.SetLocationCommand;
import net.astah.dd.di.edit.provider.LayoutCalculator;
import net.astah.dd.di.util.DiagramElements;
import net.astah.golf.util.ResourceBundleList;
import net.astah.jomt.jsystem.PropertyManager;
import net.astah.notation.AnchorConstraint;
import net.astah.notation.Location;
import net.astah.notation.NotationFactory;
import net.astah.stpa.stamp.Component;
import net.astah.stpa.stamp.Link;
import net.astah.stpa.stamp.LinkEnd;
import net.astah.stpa.stamp.notation.STAMPComponentShape;
import net.astah.stpa.stamp.notation.STAMPEdge;
import net.astah.stpa.stamp.notation.STAMPShape;

/**
 * コントロールストラクチャー図に図要素を追加した際に、図要素のレイアウト ( 位置、大きさの調整 ) を行うコマンドです。
 */
public class ControlStructureDiagramLayoutCommand extends CompoundCommand {
    private static final double COMPONENT_START_X = 0;
    private static final double COMPONENT_START_Y = 10;
    private static final double COMPONENT_Y_OFFSET = 50;
    private static final double EDGE_X_OFFSET = 0.05;
    private static final double EDGE_X_CENTER = 0.5;
    private static final double EDGE_Y_CENTER = 0.5;
    private static final double LINKEND_X_OFFSET = 150;
    private static final double LINKEND_Y_OFFSET = 7;
    private static final double LINKEND_DEFAULT_WIDTH = 7;
    private static final double LINKEND_DEFAULT_HEIGHT = 7;

    private final List<DiagramElement> targets = new ArrayList<>();
    private EditingDomain domain;
    private Diagram diagram;
    private Point2D point;
    private Location baseLocation;

    /**
     * レイアウト対象の図要素とレイアウトに必要な情報を受け取る.
     * 
     * @param domain
     *            EditingDomain
     * @param diagram
     *            図要素のある図
     * @param targets
     *            レイアウト対象の図要素
     */
    public ControlStructureDiagramLayoutCommand(final EditingDomain domain, final Diagram diagram,
            final List<DiagramElement> targets) {
        this(domain, diagram, targets, null);
    }

    /**
     * レイアウト対象の図要素と図要素を配置する位置、レイアウトに必要な情報を受け取る.
     * 
     * @param domain
     *            EditingDomain
     * @param diagram
     *            図要素のある図
     * @param targets
     *            レイアウト対象の図要素
     * @param baseLocation
     *            図要素を配置する位置
     */
    public ControlStructureDiagramLayoutCommand(final EditingDomain domain, final Diagram diagram,
            final List<DiagramElement> targets, final Location baseLocation) {
        super();
        this.domain = domain;
        this.diagram = diagram;
        this.targets.addAll(targets);
        this.baseLocation = baseLocation;
    }

    /**
     * 全てのコマンドを実行可能か返します。 これにより、 {@link #isExecutable} でキャッシュすることができます。
     * コマンドリストがからの場合には <code>false</code> を返します。
     * 
     * @return 全てのコマンドが実行可能であれば true 、 1 つでも実行不可であれば false
     */
    @Override
    protected boolean prepare() {
        createSetLayoutCommand(baseLocation);
        return super.prepare();
    }

    private void createSetLayoutCommand(final Location location) {
        final LayoutCalculator layoutCalculator = new LayoutCalculator();
        final DiagramContext context = new DiagramContext(null);
        final double defaultWidth = getDefaultWidth();
        final double defaultHeight = getDefaultHeight();
        double margin = 0;
        final List<DiagramElement> linkEnds = targets.stream()
                .filter(target -> isLinkEnd(target)).collect(Collectors.toList());
        final List<DiagramElement> links = targets.stream()
                .filter(target -> target instanceof STAMPEdge).collect(Collectors.toList());
        for (DiagramElement target : targets) {
            if (target instanceof STAMPComponentShape) {
                final STAMPShape shape = (STAMPShape) target;
                final Point value = getNewPoint(location, layoutCalculator, margin, shape);
                append(SetLocationCommand.create(domain, context, shape, value));
                margin += defaultHeight + COMPONENT_Y_OFFSET;

                // 自分と関係するLinkEndがあればそれもレイアウト設定する
                final Bounds bounds = DCFactory.eINSTANCE
                        .createBounds(value.getX(), value.getY(), defaultWidth, defaultHeight);
                double inputMargin = 0;
                double outputMargin = 0;
                for (DiagramElement element : linkEnds) {
                    final STAMPShape linkEnd = (STAMPShape) element;
                    final boolean isInput = links.stream()
                            .anyMatch(link -> isInput((STAMPEdge) link, linkEnd, shape));
                    final boolean isOutput = links.stream()
                            .anyMatch(link -> isOutput((STAMPEdge) link, linkEnd, shape));

                    if (isInput) {
                        // 入力のLinkEndのレイアウト
                        final Point point = createSetInputLinkEndPoint(bounds, inputMargin);
                        inputMargin += LINKEND_DEFAULT_HEIGHT + LINKEND_Y_OFFSET;
                        append(SetLocationCommand.create(domain, context, linkEnd, point));
                    } else if (isOutput) {
                        // 出力のLinkEndのレイアウト
                        final Point point = createSetOutputLinkEndPoint(bounds, outputMargin);
                        outputMargin += LINKEND_DEFAULT_HEIGHT + LINKEND_Y_OFFSET;
                        append(SetLocationCommand.create(domain, context, linkEnd, point));
                    }
                }
            } else if (target instanceof STAMPEdge) {
                setEdgeAnchors((Edge) target);
            } else if (isLinkEnd(target)) {
                final STAMPShape linkEnd = (STAMPShape) target;
                final Optional<STAMPEdge> sourceEdge = linkEnd.getSTAMPSourceEdges().stream()
                        .filter(edge -> !targets.contains(edge.getTarget()))
                        .findFirst();
                final Optional<STAMPEdge> targetEdge = linkEnd.getSTAMPTargetEdges().stream()
                        .filter(edge -> !targets.contains(edge.getSource()))
                        .findFirst();

                if (sourceEdge.isPresent()) {
                    // 入力のLinkEndのレイアウト
                    final Bounds bounds = ((STAMPShape) sourceEdge.get().getTarget()).getBounds();
                    final Point point = createSetInputLinkEndPoint(bounds, 0.0);
                    append(SetLocationCommand.create(domain, context, linkEnd, point));
                } else if (targetEdge.isPresent()) {
                    // 出力のLinkEndのレイアウト
                    final Bounds bounds = ((STAMPShape) targetEdge.get().getSource()).getBounds();
                    final Point point = createSetOutputLinkEndPoint(bounds, 0.0);
                    append(SetLocationCommand.create(domain, context, linkEnd, point));
                }
            }
        }
    }

    private Point createSetInputLinkEndPoint(final Bounds bounds, final double margin) {
        final double linkEndX = bounds.getX() - LINKEND_X_OFFSET - LINKEND_DEFAULT_WIDTH / 2;
        final double linkEndY = bounds.getY() + bounds.getHeight() / 2 - LINKEND_DEFAULT_HEIGHT / 2
                + margin;
        return DCFactory.eINSTANCE.createPoint(linkEndX, linkEndY);

    }

    private Point createSetOutputLinkEndPoint(final Bounds bounds, final double margin) {
        final double rightX = bounds.getX() + bounds.getWidth();
        final double linkEndX = rightX + LINKEND_X_OFFSET - LINKEND_DEFAULT_WIDTH / 2;
        final double linkEndY = bounds.getY() + bounds.getHeight() / 2 - LINKEND_DEFAULT_HEIGHT / 2
                + margin;
        return DCFactory.eINSTANCE.createPoint(linkEndX, linkEndY);
    }

    private boolean isInput(final STAMPEdge edge, final STAMPShape linkEnd,
            final STAMPShape component) {
        final Link linkModel = (Link) DiagramElements.getModelElement(edge);
        final LinkEnd linkEndModel = (LinkEnd) DiagramElements.getModelElement(linkEnd);
        final Component componentModel = (Component) DiagramElements.getModelElement(component);
        return linkModel.getSource() == linkEndModel && linkModel.getTarget() == componentModel;
    }

    private boolean isOutput(final STAMPEdge edge, final STAMPShape linkEnd,
            final STAMPShape component) {
        final Link linkModel = (Link) DiagramElements.getModelElement(edge);
        final LinkEnd linkEndModel = (LinkEnd) DiagramElements.getModelElement(linkEnd);
        final Component componentModel = (Component) DiagramElements.getModelElement(component);
        return linkModel.getSource() == componentModel && linkModel.getTarget() == linkEndModel;
    }

    private boolean isLinkEnd(final DiagramElement element) {
        return element instanceof STAMPShape
                && DiagramElements.getModelElement(element) instanceof LinkEnd;
    }

    private Point getNewPoint(final Location location, final LayoutCalculator layoutCalculator,
            final double margin, final STAMPShape target) {
        if (location == null) {
            final Point basePoint = getShapePoint(layoutCalculator, target);
            return DCFactory.eINSTANCE.createPoint(basePoint.getX(), basePoint.getY() + margin);
        }
        return DCFactory.eINSTANCE.createPoint(location.getX(), location.getY() + margin);
    }

    private Point getShapePoint(final LayoutCalculator layoutCalculator, final STAMPShape target) {
        final DiagramElement owningElement = target.getOwningElement();
        final Shape parent = owningElement instanceof STAMPShape ? (Shape) owningElement : diagram;
        final double bottomY = getBottomY(parent, layoutCalculator);
        return DCFactory.eINSTANCE.createPoint(COMPONENT_START_X, bottomY);
    }

    /**
     * 対象Edgeと、Edgeと同じSource,TargetのEdgeの位置を調整し、セットします。
     * 
     * @param edge
     *            対象Edge
     */
    private void setEdgeAnchors(final Edge edge) {
        final DiagramElement source = edge.getSource();
        final DiagramElement target = edge.getTarget();
        if (source == null || target == null) {
            return;
        }

        final List<Edge> siblings = findEdges(source, target);
        for (int index = 0; index < siblings.size(); index++) {
            double rx = EDGE_X_CENTER;
            if (siblings.size() != 1) {
                rx = EDGE_X_OFFSET + (1 - 2 * EDGE_X_OFFSET) * index / (siblings.size() - 1);
            }
            final AnchorConstraint sourceAnchor = NotationFactory.eINSTANCE.createRatioAnchor(rx,
                    EDGE_Y_CENTER);
            final AnchorConstraint targetAnchor = NotationFactory.eINSTANCE.createRatioAnchor(rx,
                    EDGE_Y_CENTER);
            append(createDiagramCommand(domain, SetEdgeEndCommand.class,
                    new CommandParameter(edge, SetEdgeEndCommand.SOURCE_ANCHOR, sourceAnchor)));
            append(createDiagramCommand(domain, SetEdgeEndCommand.class,
                    new CommandParameter(edge, SetEdgeEndCommand.TARGET_ANCHOR, targetAnchor)));
        }
    }

    /**
     * 2つのDiagramElement間のEdgeを取得します。
     * 
     * @param end1
     *            DiagramElement
     * @param end2
     *            DiagramElement
     * @return DiagramElement間のEdge
     */
    private List<Edge> findEdges(final DiagramElement end1, final DiagramElement end2) {
        final List<Edge> edges = new ArrayList<>();
        edges.addAll(Collections2.filter(end1.getSourceEdges(), edge -> edge.getTarget() == end2));
        edges.addAll(Collections2.filter(end1.getTargetEdges(), edge -> edge.getSource() == end2));
        return edges;
    }

    /**
     * コンポーネントの図要素配置時のY座標を取得します。
     * 
     * @param owningElement
     * @param layoutCalculator
     * @return 配置Y座標
     */
    private double getBottomY(final DiagramElement owningElement,
            final LayoutCalculator layoutCalculator) {
        double bottomOfChilren = point == null ? COMPONENT_START_Y : point.getY();

        final List<DiagramElement> parentOwnedElements = owningElement.getOwnedElements().stream()
                .filter(element -> !targets.contains(element)).collect(Collectors.toList());
        if (parentOwnedElements.isEmpty()) {
            return bottomOfChilren;
        }

        final Point origin = layoutCalculator.getBoundsLocation(owningElement);
        for (DiagramElement child : owningElement.getOwnedElements()) {
            final Bounds bounds = layoutCalculator.getBounds(child);
            final Bounds relativize = LayoutCalculator.relativize(bounds, origin);
            final double maxY = relativize.getY() + relativize.getHeight();
            bottomOfChilren = Math.max(bottomOfChilren, maxY);
        }
        if (point != null) {
            point.setLocation(point.getX(), bottomOfChilren + COMPONENT_Y_OFFSET);
        }
        return bottomOfChilren + COMPONENT_Y_OFFSET;
    }

    private double getDefaultHeight() {
        final ResourceBundleList resouces = PropertyManager.getJudeResourceBundle();
        if (resouces != null) {
            final String strHeight = resouces.getString("default_size.component.default_height");
            return Double.parseDouble(strHeight);
        }
        return 80;
    }

    private double getDefaultWidth() {
        final ResourceBundleList resouces = PropertyManager.getJudeResourceBundle();
        if (resouces != null) {
            final String strHeight = resouces.getString("default_size.component.default_width");
            return Double.parseDouble(strHeight);
        }
        return 180;
    }

}
