
package net.astah.stpa.stamp.ui.control.mode;

import static java.util.Collections.singleton;

import java.awt.dnd.DropTargetDropEvent;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import org.eclipse.emf.common.command.CompoundCommand;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.edit.command.CommandParameter;
import org.eclipse.emf.edit.command.CreateChildCommand;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.emf.transaction.util.TransactionUtil;

import com.google.common.collect.Lists;

import net.astah.dd.dc.DCFactory;
import net.astah.dd.di.DIPackage;
import net.astah.dd.di.Diagram;
import net.astah.dd.di.DiagramElement;
import net.astah.dd.di.Shape;
import net.astah.dd.di.util.DiagramElements;
import net.astah.emf.edit.job.CommandJob;
import net.astah.golf.geom2D.Pnt2d;
import net.astah.golf.model.IEntityStore;
import net.astah.golf.widget.swing.Selections;
import net.astah.jomt.jsystem.JomtGlobal;
import net.astah.jomt.jsystem.ModelGlobal;
import net.astah.jomt.jview.ProjectView;
import net.astah.notation.Location;
import net.astah.notation.NotationFactory;
import net.astah.stpa.stamp.Component;
import net.astah.stpa.stamp.ControlStructure;
import net.astah.stpa.stamp.Link;
import net.astah.stpa.stamp.LinkEnd;
import net.astah.stpa.stamp.LinkableElement;
import net.astah.stpa.stamp.STPAAnalysis;
import net.astah.stpa.stamp.notation.STAMPControlLoopDiagram;
import net.astah.stpa.stamp.notation.STAMPControlStructureDiagram;
import net.astah.stpa.stamp.notation.STAMPEdge;
import net.astah.stpa.stamp.notation.STAMPNotationFactory;
import net.astah.stpa.stamp.notation.STAMPShape;
import net.astah.stpa.stamp.notation.util.STAMPDiagramElements;
import net.astah.stpa.stamp.ui.control.ControlStructureDiagramLayoutCommand;
import net.astah.ui.control.CreateEdgeCommand;
import net.astah.ui.control.CreateShapeCommand;
import net.astah.ui.control.mode.DiagramDnDMode;

/**
 * 構造ツリーからコントロールストラクチャー図へモデルをドラッグ & ドロップした場合の処理を行う Mode です。
 */
public class ControlStructureDiagramDnDMode extends DiagramDnDMode {

    public ControlStructureDiagramDnDMode() {
        super();
    }

    @Override
    protected void dropExecute(DropTargetDropEvent dtde) {
        IEntityStore doc = ModelGlobal.prjManager.getCurrentProject().doc;
        if (doc == null) {
            return;
        }
        TransactionalEditingDomain domain = (TransactionalEditingDomain) doc.getDomain();
        boolean isInTransaction = doc.isInTransaction();
        if (isInTransaction) {
            return;
        }

        Pnt2d p = getLocation(dtde);
        final Location location = NotationFactory.eINSTANCE.createLocation(p.getX(), p.getY());
        ProjectView pv = JomtGlobal.viewManager.getProjectView();

        final Diagram diagram = getDiagram();
        final ControlStructure controlStructure = getControlStructure(diagram);
        if (controlStructure == null) {
            return;
        }

        final CompoundCommand command = new CompoundCommand();

        // 新規に作成した図要素のみレイアウトの対象とする
        final List<DiagramElement> layoutTargets = new ArrayList<>();
        final List<Component> components = Lists
                .newArrayList(Selections.iterable(pv.getSelection(), Component.class));
        final EList<Link> links = controlStructure.getLinks();

        for (Component component : components) {
            // ControlStructureにComponentのShapeを追加する
            createComponentShapeCommand(component, diagram, layoutTargets, command);
            createLinkEnds(domain, links, component, diagram, layoutTargets, command);
        }
        createEdges(domain, links, layoutTargets, diagram, command);
        createLayoutCommand(domain, layoutTargets, location, diagram, command);
        if (!command.canExecute()) {
            return;
        }

        final CommandJob job = new CommandJob(domain, command);
        job.schedule();
    }

    private ControlStructure getControlStructure(Diagram diagram) {
        if (diagram instanceof STAMPControlStructureDiagram) {
            return DiagramElements.getModelElement(diagram, ControlStructure.class);
        } else if (diagram instanceof STAMPControlLoopDiagram) {
            final STPAAnalysis analysis = ((STAMPControlLoopDiagram) diagram).getControlLoop()
                    .getHintWordSet().getOwner();
            if (!analysis.getControlStructures().isEmpty()) {
                return analysis.getControlStructures().get(0);
            }
        }
        return null;
    }

    private void createComponentShapeCommand(final Component component, final Diagram diagram,
            final List<DiagramElement> layoutTargets, final CompoundCommand command) {
        Shape parent = STAMPDiagramElements
                .getDiagramElements(component.eContainer(), diagram, STAMPShape.class)
                .stream().map(m -> (Shape) m).findFirst().orElse(diagram);

        final STAMPShape shape = STAMPNotationFactory.eINSTANCE.createSTAMPComponentShape();
        // 図要素の登録に先立ってモデル要素を追加しておく必要がある
        DiagramElements.setModelElement(shape, component);
        shape.getModelElements().add(component);

        // 位置設定はControlStructureDiagramLayoutCommandで行う
        shape.setLocation(NotationFactory.eINSTANCE.createLocation(0, 0));
        layoutTargets.add(shape);
        TransactionalEditingDomain domain = TransactionUtil.getEditingDomain(parent);

        command.append(CreateChildCommand.create(domain, parent,
                new CommandParameter(parent,
                        DIPackage.Literals.DIAGRAM_ELEMENT__OWNED_ELEMENT, shape, 0),
                singleton(parent)));
    }

    /**
     * 始点・終点の図要素を作成します。
     */
    private void createLinkEnds(final TransactionalEditingDomain domain,
            final EList<Link> links, final Component component, final Diagram diagram,
            final List<DiagramElement> layoutTargets, final CompoundCommand command) {
        // 始点の作成
        final List<Link> inputLinks = links.stream()
                .filter(link -> link.getTarget() == component
                        && link.getSource() instanceof LinkEnd)
                .collect(Collectors.toList());
        for (Link inputLink : inputLinks) {
            final LinkEnd linkEnd = (LinkEnd) inputLink.getSource();
            if (isExistLinkEnd(diagram, layoutTargets, linkEnd)) {
                continue;
            }

            final STAMPShape shape = STAMPNotationFactory.eINSTANCE.createSTAMPShape();
            shape.getModelElements().add(linkEnd);

            final CreateShapeCommand cmd = createCreateShapeCommand(diagram, shape);
            command.append(cmd.createTransactionalCommand(domain));
            layoutTargets.add(shape);
        }

        // 終点の作成
        List<Link> outputLinks = links.stream()
                .filter(link -> link.getSource() == component
                        && link.getTarget() instanceof LinkEnd)
                .collect(Collectors.toList());
        for (Link inputLink : outputLinks) {
            final LinkEnd linkEnd = (LinkEnd) inputLink.getTarget();
            if (isExistLinkEnd(diagram, layoutTargets, linkEnd)) {
                continue;
            }

            final STAMPShape shape = STAMPNotationFactory.eINSTANCE.createSTAMPShape();
            shape.getModelElements().add(linkEnd);

            final CreateShapeCommand cmd = createCreateShapeCommand(diagram, shape);
            command.append(cmd.createTransactionalCommand(domain));
            layoutTargets.add(shape);
        }
    }

    private boolean isExistLinkEnd(final Diagram diagram, final List<DiagramElement> layoutTargets,
            final LinkEnd linkEnd) {
        final boolean isExistInTargets = layoutTargets.stream()
                .anyMatch(target -> DiagramElements.getModelElement(target) == linkEnd);
        final boolean isExistInDiagram = diagram.getOwnedElements().stream()
                .anyMatch(element -> DiagramElements.getModelElement(element) == linkEnd);
        return isExistInTargets || isExistInDiagram;
    }

    private CreateShapeCommand createCreateShapeCommand(final DiagramElement parent,
            final STAMPShape shape) {
        CreateShapeCommand cmd = new CreateShapeCommand();
        cmd.setOwningElement(parent);
        cmd.setShape(shape);
        cmd.setLocation(DCFactory.eINSTANCE.createPoint(0, 0));
        return cmd;
    }

    private void createLayoutCommand(final TransactionalEditingDomain domain,
            final List<DiagramElement> layoutTargets, final Location location,
            final Diagram diagram, final CompoundCommand command) {
        command.append(
                new ControlStructureDiagramLayoutCommand(domain, diagram, layoutTargets, location));
    }

    @SuppressWarnings("deprecation")
    private void createEdges(final TransactionalEditingDomain domain,
            final EList<Link> links, final List<DiagramElement> layoutTargets,
            final Diagram diagram,
            final CompoundCommand command) {
        for (Link link : links) {

            STAMPShape sourceShape = getRelationShape(link.getSource(), layoutTargets, diagram);
            STAMPShape targetShape = getRelationShape(link.getTarget(), layoutTargets, diagram);
            if (targetShape == null || sourceShape == null) {
                continue;
            }

            Optional<STAMPEdge> linkEdge = STAMPDiagramElements
                    .getDiagramElements(link, diagram, STAMPEdge.class).stream().findFirst();
            if (linkEdge.equals(Optional.empty())) {
                final CreateEdgeCommand createEdgeCommand = new CreateEdgeCommand();
                final STAMPEdge edge = STAMPNotationFactory.eINSTANCE.createSTAMPEdge();
                edge.getModelElements().add(link);
                createEdgeCommand.setEdge(edge);
                createEdgeCommand.setSource(sourceShape);
                createEdgeCommand.setTarget(targetShape);
                // DnDされたコンポーネントに接続するEdgeは図直下に作成する
                createEdgeCommand.setOwningElement(diagram);
                createEdgeCommand.setLine(NotationFactory.eINSTANCE.createRightAngleLine());
                command.append(createEdgeCommand.createTransactionalCommand(domain));
                layoutTargets.add(edge);
            }
        }
    }

    private STAMPShape getRelationShape(final LinkableElement linkableElement,
            final List<DiagramElement> layoutTargets, final Diagram diagram) {
        final Optional<DiagramElement> optionalElement = layoutTargets.stream()
                .filter(layoutTarget -> DiagramElements
                        .getModelElement(layoutTarget) == linkableElement)
                .findFirst();
        if (optionalElement.isPresent()) {
            final DiagramElement element = optionalElement.get();
            if (element instanceof STAMPShape) {
                return (STAMPShape) element;
            }
        }

        return STAMPDiagramElements
                .getDiagramElements(linkableElement, diagram, STAMPShape.class).stream()
                .findFirst().orElse(null);
    }
}