package net.astah.stpa.stamp.ui.menu.handler;

import static net.astah.stpa.stamp.STAMPPackage.Literals.HAZARD_CAUSAL_FACTOR__COUNTERMEASURE;
import static org.eclipse.emf.edit.domain.AdapterFactoryEditingDomain.getEditingDomainFor;

import java.awt.event.ActionEvent;
import java.util.Comparator;
import java.util.List;
import java.util.stream.Collectors;

import javax.swing.JFrame;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.command.CompoundCommand;
import org.eclipse.emf.common.command.UnexecutableCommand;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.edit.command.AddCommand;
import org.eclipse.emf.edit.command.RemoveCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.emf.transaction.util.TransactionUtil;

import net.astah.emf.edit.command.DestroyCommand;
import net.astah.emf.edit.job.CommandJob;
import net.astah.golf.widget.IListSelection;
import net.astah.golf.widget.ISelection;
import net.astah.golf.widget.ITableSelection;
import net.astah.golf.widget.swing.SwingFrame;
import net.astah.jomt.jsystem.JomtGlobal;
import net.astah.jomt.jview.Editor;
import net.astah.jomt.jview.EditorPeer;
import net.astah.stpa.stamp.Countermeasure;
import net.astah.stpa.stamp.HazardCausalFactor;
import net.astah.stpa.stamp.IdentifiedElement;
import net.astah.stpa.stamp.STPAAnalysis;
import net.astah.stpa.stamp.ui.Messages;
import net.astah.stpa.stamp.ui.view.SelectElementDialog;
import net.astah.stpa.stamp.ui.view.swing.CountermeasureTableEditorPeer;
import net.astah.table.Element;
import net.astah.table.Tuple;
import net.astah.ui.menu.annotation.CanExecute;
import net.astah.ui.menu.annotation.Execute;
import net.astah.ui.menu.annotation.Handler;
import net.astah.ui.menu.annotation.MenuCommand;
import net.astah.ui.menu.annotation.MenuItem;
import net.astah.ui.menu.annotation.MenuItem.ItemKind;

/**
 * 対策表で右クリックした対策のセルに設定する対策を、対策選択ダイアログで指定します。 <br>
 * 同じ HCF-UCA の組み合わせに同じ対策を複数指定することはできません。
 */
@Handler
@MenuItem(commands = @MenuCommand(id = "stpa.SelectCountermeasureCommand",
        category = "edit.table"),
        label = "%StampTable_selectCountermeasure_label",
        tooltip = "%StampTable_selectCountermeasure_tooltip",
        type = ItemKind.PUSH)
public class SelectCountermeasureHandler {

    private int COUNTERMEASURE_ID_COLUMN = 2;
    private int COUNTERMEASURE_COLUMN = 3;

    private Command command = UnexecutableCommand.INSTANCE;

    /**
     * 対策選択ダイアログで指定した対策を設定するコマンドを実行します。
     * 
     * @param selection
     *            選択要素
     * @param params
     *            コマンドのパラメータ
     * @param e
     *            アクションイベント
     */
    @Execute
    public void execute(final ISelection selection, final List<String> params,
            final ActionEvent e) {
        if (selection instanceof ITableSelection) {
            final ITableSelection tableSelection = (ITableSelection) selection;
            command = selectCountermeasure((Tuple) tableSelection.getFirstElement());
            final CommandJob job = new CommandJob(getDomain(), command);
            job.schedule();
        }
    }

    private TransactionalEditingDomain getDomain() {
        final Editor editor = JomtGlobal.viewManager.getCurrentEditor();
        return TransactionUtil.getEditingDomain(editor.getEditModel());
    }

    /**
     * 対策表の右クリックしたセルで対策を選択できるかどうか確認します。
     * 
     * @param selection
     *            選択要素
     * @return true 右クリック箇所のセルに設定可能な対策が存在する場合 / false 上記以外
     */
    @CanExecute
    public boolean canExecute(final ISelection selection) {
        return selection instanceof ITableSelection && isTarget((ITableSelection) selection);
    }

    private boolean isTarget(final ITableSelection selection) {
        final Object element = selection.getFirstElement();
        if (element instanceof Tuple) {
            return existCountermeasures((Tuple) element)
                    && isCountermeasure(selection.getColumnSelection());
        }
        return false;
    }

    private boolean existCountermeasures(Tuple rowTuple) {
        final CountermeasureTableEditorPeer editorPeer = getEditorPeer();
        final STPAAnalysis scope = editorPeer.getScope();
        final HazardCausalFactor hcf = (HazardCausalFactor) rowTuple.getElements()
                .get(CountermeasureTableEditorPeer.HAZARD_CAUSAL_FACTOR_INDEX).getObject();
        final Countermeasure countermeasure = (Countermeasure) rowTuple.getElements()
                .get(CountermeasureTableEditorPeer.COUNTERMEASURE_INDEX).getObject();

        return scope.getCountermeasures().stream().anyMatch(m -> m != countermeasure
                && !m.getHazardCausalFactors().contains(hcf));
    }

    private boolean isCountermeasure(final IListSelection columnSelection) {
        final int[] indices = columnSelection.getIndices();
        if (indices.length > 0) {
            return indices[0] == COUNTERMEASURE_COLUMN || indices[0] == COUNTERMEASURE_ID_COLUMN;
        }
        return false;
    }

    private Command selectCountermeasure(final Tuple rowTuple) {
        final CountermeasureTableEditorPeer editorPeer = getEditorPeer();
        final HazardCausalFactor hcf = (HazardCausalFactor) rowTuple.getElements()
                .get(CountermeasureTableEditorPeer.HAZARD_CAUSAL_FACTOR_INDEX).getObject();
        final Countermeasure countermeasure = (Countermeasure) rowTuple.getElements()
                .get(CountermeasureTableEditorPeer.COUNTERMEASURE_INDEX).getObject();
        final JFrame parent = (JFrame) ((SwingFrame) JomtGlobal.viewManager.getManagementView()
                .getComponent()).getComponent();

        final List<IdentifiedElement> list = editorPeer.getScope().getCountermeasures().stream()
                .filter(m -> m == countermeasure || !m.getHazardCausalFactors().contains(hcf))
                .sorted(Comparator.comparing(IdentifiedElement::getId, String.CASE_INSENSITIVE_ORDER))
                .collect(Collectors.toList());
        final SelectElementDialog dialog = new SelectElementDialog(parent, list, countermeasure);
        dialog.setTitle(Messages.countermeasure_table_select_countermeasure_label);
        dialog.pack();
        dialog.setVisible(true);

        if (dialog.getResultState() == SelectElementDialog.OK) {
            return createSetCountermeasureCommand(rowTuple,
                    (Countermeasure) dialog.getSelectedElement(), editorPeer);
        }
        return command;
    }

    private Command createSetCountermeasureCommand(final Tuple rowTuple,
            final Countermeasure countermeasure, final CountermeasureTableEditorPeer editorPeer) {
        final STPAAnalysis scope = editorPeer.getScope();
        final EditingDomain domain = getEditingDomainFor(scope);
        final EList<Element> elements = rowTuple.getElements();
        final Countermeasure oldCountermeasure = (Countermeasure) elements
                .get(CountermeasureTableEditorPeer.COUNTERMEASURE_INDEX).getObject();
        final CompoundCommand command = new CompoundCommand();

        if (oldCountermeasure != null) {
            if (oldCountermeasure != countermeasure) {
                if (oldCountermeasure.getHazardCausalFactors().size() == 1) {
                    command.append(DestroyCommand.create(domain, oldCountermeasure));
                } else {
                    command.append(RemoveCommand.create(domain,
                            rowTuple.getElements()
                                    .get(CountermeasureTableEditorPeer.HAZARD_CAUSAL_FACTOR_INDEX)
                                    .getObject(),
                            HAZARD_CAUSAL_FACTOR__COUNTERMEASURE, oldCountermeasure));
                }
            } else {
                return UnexecutableCommand.INSTANCE;
            }
        }

        command.append(AddCommand.create(domain,
                rowTuple.getElements().get(CountermeasureTableEditorPeer.HAZARD_CAUSAL_FACTOR_INDEX)
                        .getObject(),
                HAZARD_CAUSAL_FACTOR__COUNTERMEASURE, countermeasure));

        return command;
    }

    private CountermeasureTableEditorPeer getEditorPeer() {
        final Editor editor = JomtGlobal.viewManager.getCurrentEditor();
        final EditorPeer editorPeer = editor.getEditorPeer();
        return (CountermeasureTableEditorPeer) editorPeer;
    }

}
